#!/usr/bin/env python3
"""Service account module."""

import os
import subprocess
from argparse import ArgumentParser, Namespace
from enum import Enum
from logging import Logger

from lightkube import ApiError

from spark8t.cli import defaults
from spark8t.cli.params import (
    add_config_arguments,
    add_logging_arguments,
    k8s_parser,
    parse_arguments_with,
    spark_user_parser,
)
from spark8t.domain import KubernetesResourceType, ServiceAccount
from spark8t.exceptions import (
    AccountNotFound,
    NamespaceNotFound,
    PrimaryAccountNotFound,
    ResourceAlreadyExists,
)
from spark8t.kube_interface.base import AbstractKubeInterface
from spark8t.kube_interface.lightkube import LightKubeInterface
from spark8t.registry.k8s import K8sServiceAccountRegistry
from spark8t.utils import PropertyFile, setup_logging


def build_service_account_from_args(args, registry) -> ServiceAccount:
    """Create service account resource interface."""
    return ServiceAccount(
        name=os.path.expandvars(args.username),
        namespace=os.path.expandvars(args.namespace),
        api_server=registry.kube_interface.api_server,
        primary=args.primary if hasattr(args, "primary") else False,
    )


class Actions(str, Enum):
    """Service account CLI action."""

    CREATE = "create"
    DELETE = "delete"
    ADD_CONFIG = "add-config"
    REMOVE_CONFIG = "remove-config"
    GET_CONFIG = "get-config"
    CLEAR_CONFIG = "clear-config"
    PRIMARY = "get-primary"
    LIST = "list"
    GET_MANIFEST = "get-manifest"

    def __str__(self) -> str:
        """Define string representation.

        TODO(py310): replace inheritance with StrEnum once we drop py310
        """
        return str.__str__(self)


def create_namespace_if_missing(kube_interface: AbstractKubeInterface, namespace: str):
    """Create namespace if does not exist."""
    if not kube_interface.exists(KubernetesResourceType.NAMESPACE, namespace):
        try:
            kube_interface.create(KubernetesResourceType.NAMESPACE, namespace)
        except ApiError as e:
            if e.status.code == 401 or e.status.code == 403:
                print(f"Namespace {namespace} can not be created.")
                raise NamespaceNotFound(namespace) from None
            else:
                raise e
        except subprocess.CalledProcessError as err:
            print(f"Namespace {namespace} can not be created.")
            raise NamespaceNotFound(namespace) from err


def create_service_account_registry_parser(parser: ArgumentParser):
    """Create parser for service account CLI."""
    base_parser = parse_arguments_with(
        [add_logging_arguments, k8s_parser],
        ArgumentParser(add_help=False),
    )

    subparsers = parser.add_subparsers(dest="action")
    subparsers.required = True

    #  subparser for service-account
    parse_arguments_with(
        [add_config_arguments, spark_user_parser],
        subparsers.add_parser(Actions.CREATE.value, parents=[base_parser]),
    ).add_argument(
        "--primary",
        action="store_true",
        help="Boolean to mark the service account as primary.",
    )

    #  subparser for service-account-cleanup
    parse_arguments_with(
        [spark_user_parser],
        subparsers.add_parser(Actions.DELETE.value, parents=[base_parser]),
    )

    #  subparser for add-config
    parse_arguments_with(
        [add_config_arguments, spark_user_parser],
        subparsers.add_parser(Actions.ADD_CONFIG.value, parents=[base_parser]),
    )

    #  subparser for remove-config
    parse_arguments_with(
        [add_config_arguments, spark_user_parser],
        subparsers.add_parser(Actions.REMOVE_CONFIG.value, parents=[base_parser]),
    )

    #  subparser for sa-conf-get
    parse_arguments_with(
        [spark_user_parser],
        subparsers.add_parser(Actions.GET_CONFIG.value, parents=[base_parser]),
    ).add_argument(
        "--ignore-integration-hub",
        action="store_true",
        help="Boolean to ignore Spark Integration Hub generated options.",
    )

    #  subparser for get-manifest
    parse_arguments_with(
        [spark_user_parser],
        subparsers.add_parser(Actions.GET_MANIFEST.value, parents=[base_parser]),
    )

    #  subparser for sa-conf-del
    parse_arguments_with(
        [spark_user_parser],
        subparsers.add_parser(Actions.CLEAR_CONFIG.value, parents=[base_parser]),
    )

    #  subparser for resources-primary-sa
    subparsers.add_parser(Actions.PRIMARY.value, parents=[base_parser])

    #  subparser for list
    subparsers.add_parser(Actions.LIST.value, parents=[base_parser])

    return parser


def main(args: Namespace, logger: Logger):
    """Service account main entrypoint."""
    kubeconfig = os.path.expandvars(args.kubeconfig) if args.kubeconfig else None
    context_name = os.path.expandvars(args.context) if args.context else None

    kube_interface = LightKubeInterface(
        kubeconfig or defaults.kube_config, defaults, context_name=context_name
    )
    context = context_name or kube_interface.context_name

    logger.debug(f"Using K8s context: {context}")

    registry = K8sServiceAccountRegistry(
        kube_interface.with_context(context) if context else kube_interface
    )

    if args.action == Actions.CREATE:
        service_account = build_service_account_from_args(args, registry)

        # check if namespace exist otherwise create it if permissions allow it.
        create_namespace_if_missing(kube_interface, service_account.namespace)
        properties_file = (
            os.path.expandvars(args.properties_file) if args.properties_file else None
        )
        service_account.extra_confs = (
            PropertyFile.read(properties_file)
            if properties_file is not None
            else PropertyFile.empty()
        ) + PropertyFile.parse_conf_overrides(args.conf)
        registry.create(service_account)

    elif args.action == Actions.GET_MANIFEST:
        service_account = build_service_account_from_args(args, registry)
        manifest = registry.create(service_account, dry_run=True)
        print(manifest)

    elif args.action == Actions.DELETE:
        user_id = build_service_account_from_args(args, registry).id
        logger.info(user_id)
        registry.delete(user_id)

    elif args.action == Actions.ADD_CONFIG:
        input_service_account = build_service_account_from_args(args, registry)

        service_account_in_registry = registry.get(input_service_account.id)

        if service_account_in_registry is None:
            raise AccountNotFound(input_service_account.id)
        properties_file = (
            os.path.expandvars(args.properties_file) if args.properties_file else None
        )
        account_configuration = (
            service_account_in_registry.extra_confs
            + (
                PropertyFile.read(properties_file)
                if properties_file is not None
                else PropertyFile.empty()
            )
            + PropertyFile.parse_conf_overrides(args.conf)
        )

        registry.set_configurations(input_service_account.id, account_configuration)

    elif args.action == Actions.REMOVE_CONFIG:
        input_service_account = build_service_account_from_args(args, registry)

        service_account_in_registry = registry.get(input_service_account.id)

        if service_account_in_registry is None:
            raise AccountNotFound(input_service_account.id)

        registry.set_configurations(
            input_service_account.id,
            service_account_in_registry.extra_confs.remove(args.conf),
        )

    elif args.action == Actions.GET_CONFIG:
        input_service_account = build_service_account_from_args(args, registry)

        maybe_service_account = registry.get(input_service_account.id)

        if maybe_service_account is None:
            raise AccountNotFound(input_service_account.id)

        if args.ignore_integration_hub:
            maybe_service_account.integration_hub_confs = PropertyFile.empty()
        maybe_service_account.configurations.log(print)

    elif args.action == Actions.CLEAR_CONFIG:
        registry.set_configurations(
            build_service_account_from_args(args, registry).id, PropertyFile.empty()
        )

    elif args.action == Actions.PRIMARY:
        maybe_service_account = registry.get_primary()

        if maybe_service_account is None:
            raise PrimaryAccountNotFound()

        print(maybe_service_account.id)

    elif args.action == Actions.LIST:
        for service_account in registry.all():
            print_line = f"{service_account.id}"
            if service_account.primary:
                print_line += " (Primary)"
            print(print_line)


if __name__ == "__main__":
    args = create_service_account_registry_parser(
        ArgumentParser(description="Spark Client Setup")
    ).parse_args()

    logger = setup_logging(
        args.log_level, args.log_conf_file, "spark8t.cli.service_account_registry"
    )

    try:
        main(args, logger)
        exit(0)
    except (AccountNotFound, PrimaryAccountNotFound, ResourceAlreadyExists) as e:
        print(str(e))
        exit(1)
    except Exception as e:
        raise e



#!/usr/bin/env python3
# Copyright 2026 Canonical Limited
# See LICENSE file for licensing details.

"""Spark session module."""

import os
import socket
from urllib.parse import urlparse

import pyspark
from lightkube import Client
from lightkube.core.exceptions import ApiError

from spark8t.kube_interface.lightkube import LightKubeInterface as LightKube
from spark8t.registry.k8s import K8sServiceAccountRegistry
from spark8t.utils import environ


class SparkSession:
    """A context manager for Spark sessions.

    The spark session is created when entering the context and stopped when exiting. The spark session is
    created using the given namespace and service account, if provided. If not, they are inferred from the
    environment variables SPARK_NAMESPACE and SPARK_USERNAME.

    If neither are provided, a ValueError is raised.
    """

    def __init__(
        self, app_name: str, namespace: str | None = None, username: str | None = None
    ):
        self.app_name = app_name
        self.session = None

        if namespace is not None:
            self.namespace = namespace
        elif "SPARK_NAMESPACE" in os.environ:
            self.namespace = os.environ["SPARK_NAMESPACE"]
        else:
            raise ValueError(
                "Namespace must be provided either as argument or via SPARK_NAMESPACE env variable."
            )

        if username is not None:
            self.username = username
        elif "SPARK_USERNAME" in os.environ:
            self.username = os.environ["SPARK_USERNAME"]
        else:
            raise ValueError(
                "Username must be provided either as argument or via SPARK_USERNAME env variable."
            )

    @property
    def _pod_ip(
        self,
    ):
        """Return the driver pod IP address."""
        return socket.gethostbyname(socket.gethostname())

    @property
    def _sa_props(
        self,
    ):
        """Return the spark properties defined in the service account."""
        interface = LightKube(None, None)
        registry = K8sServiceAccountRegistry(interface)

        NO_PROXY = (
            os.environ.get("NO_PROXY", "") + f",{self._k8s_master_hostname}"
            if os.environ.get("NO_PROXY")
            else self._k8s_master_hostname
        )
        no_proxy = (
            os.environ.get("no_proxy", "") + f",{self._k8s_master_hostname}"
            if os.environ.get("no_proxy")
            else self._k8s_master_hostname
        )

        try:
            with environ(NO_PROXY=NO_PROXY, no_proxy=no_proxy):
                service_account = registry.get(f"{self.namespace}:{self.username}")
                return service_account.configurations.props
        except (ApiError, AttributeError):
            return {}

    @property
    def _extra_props(
        self,
    ) -> dict:
        """Return extra spark properties required for k8s communication."""
        return {"spark.driver.host": self._pod_ip}

    @property
    def _k8s_master(
        self,
    ) -> str:
        """Return the k8s api server endpoint."""
        return Client().config.cluster.server

    @property
    def _k8s_master_hostname(
        self,
    ) -> str:
        """Return the k8s api server IP address."""
        return urlparse(self._k8s_master).hostname or ""

    @property
    def config(
        self,
    ) -> dict:
        """Return the complete spark configuration dictionary."""
        return self._sa_props | self._extra_props

    def __enter__(
        self,
    ):
        """Enter the context manager, creating the spark session."""
        if self.session is not None:
            return self.session

        builder = pyspark.sql.SparkSession.builder.appName(self.app_name).master(
            f"k8s://{self._k8s_master}"
        )

        for conf, val in self.config.items():
            builder = builder.config(conf, val)
        self.session = builder.getOrCreate()
        return self.session

    def __exit__(self, *args, **kwargs):
        """Exit the context manager, stopping the spark session."""
        if self.session is not None:
            self.session.stop()



#!/usr/bin/env python3
"""Pyspark module."""

import os
import re
from argparse import Namespace
from logging import Logger

from spark8t.cli import defaults
from spark8t.cli.params import (
    add_config_arguments,
    add_ignore_integration_hub,
    add_logging_arguments,
    k8s_parser,
    parse_arguments_with,
    spark_user_parser,
)
from spark8t.domain import ServiceAccount
from spark8t.exceptions import AccountNotFound, PrimaryAccountNotFound
from spark8t.kube_interface.lightkube import LightKubeInterface
from spark8t.registry.k8s import K8sServiceAccountRegistry
from spark8t.spark_interface import SparkInterface
from spark8t.utils import PropertyFile, setup_logging


def main(args: Namespace, logger: Logger):
    """Pyspark main entrypoint."""
    kubeconfig = os.path.expandvars(args.kubeconfig) if args.kubeconfig else None
    context_name = os.path.expandvars(args.context) if args.context else None
    master = os.path.expandvars(args.master) if args.master else None
    namespace = os.path.expandvars(args.namespace) if args.namespace else None
    username = os.path.expandvars(args.username) if args.username else None
    confs = [os.path.expandvars(conf) for conf in args.conf] if args.conf else []
    properties_file = (
        os.path.expandvars(args.properties_file) if args.properties_file else None
    )

    kube_interface = LightKubeInterface(
        kubeconfig or defaults.kube_config, defaults, context_name=context_name
    )

    registry = K8sServiceAccountRegistry(
        kube_interface.select_by_master(re.compile("^k8s://").sub("", master))
        if master is not None
        else kube_interface
    )

    service_account: ServiceAccount | None = (
        registry.get_primary()
        if username is None and namespace is None
        else registry.get(f"{namespace or 'default'}:{username or 'spark'}")
    )

    if service_account is None:
        raise (AccountNotFound(username) if username else PrimaryAccountNotFound())

    if args.ignore_integration_hub:
        service_account.integration_hub_confs = PropertyFile.empty()

    SparkInterface(
        service_account=service_account,
        kube_interface=kube_interface,
        defaults=defaults,
    ).pyspark_shell(confs, properties_file, extra_args)


if __name__ == "__main__":
    args, extra_args = parse_arguments_with(
        [
            add_logging_arguments,
            k8s_parser,
            spark_user_parser,
            add_config_arguments,
            add_ignore_integration_hub,
        ]
    ).parse_known_args()

    logger = setup_logging(args.log_level, args.log_conf_file, "spark8t.cli.pyspark")

    try:
        main(args, logger)
        exit(0)
    except (AccountNotFound, PrimaryAccountNotFound) as e:
        logger.error(str(e))
        exit(1)
    except Exception as e:
        raise e
